/*
 *	Copyright (C) 2001 Daniel Gunyan <dgun@umpire.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */
/*  todo:
 *	- add transparent pixel option on command line
 *	- test on big-endian system
 *	- 16 & 32 bpp bitmaps
 *	- rle compressed bitmaps
 */

#include "toico.h"

struct bmp_info {
	long offset, header_size, bmp_size;
        int ncolors;
	int width, height;
	int bpp, compressed;
};

int isbmp(u8 *p)
{
	return *(u16*)(p) == be16(0x424D) && *(u16*)(p+6) == 0;
}

void readbmp(struct img *img, FILE *f, u8 sig[SIGN_SIZE])
{
	struct bmp_info bmp;
	u8 buf[54];
	u32 *image;

	fseek(f, -SIGN_SIZE, SEEK_CUR);
	freadn(&buf, 54, f);
	bmp.offset = get32le(&buf[10]);
	bmp.header_size = get32le(&buf[14]);
	bmp.width = get32le(&buf[18]);
	bmp.height = get32le(&buf[22]);
	bmp.bpp = get16le(&buf[28]);
	bmp.compressed = get32le(&buf[30]);
	bmp.ncolors = get32le(&buf[46]);

	switch(bmp.bpp) {
	case 1:
	case 4:
	case 8:
		if (!bmp.ncolors) bmp.ncolors = 1 << bmp.bpp;
		break;
	case 24:
		bmp.ncolors = 0;
		break;
	default:
		errx(1, "%s: Can only handle 1, 4, 8, or 24-bit bitmaps (bmp is %d-bit)",
			img->name, bmp.bpp);
	} 

	if (bmp.compressed)
		errx(1, "%s: Can't read compressed bitmaps", img->name);

	if (bmp.ncolors) {
		u32 *p, *e;
		if (bmp.ncolors > 256)
			errx(1, "%s: Too many colors", img->name);
		if (fseek(f, 14+bmp.header_size, SEEK_SET) < 0) err(1, "%s", img->name);
		freadn(img->pal, sizeof(u32) * bmp.ncolors, f);
		for(p=img->pal, e=img->pal+bmp.ncolors; p<e; p++) {
			*p = le32(*p);
			if(*p == opt.alpha_key) *p = ALPHA;
		}
	}

	bmp.bmp_size = ((((bmp.width * bmp.bpp) + 31) 
			& ~31) >> 3) * bmp.height;
	image = (u32 *)malloc(bmp.bmp_size);
	if (image == NULL) oom();
	if (fseek(f, bmp.offset, SEEK_SET) < 0) err(1, "%s", img->name);
	freadn(image, bmp.bmp_size, f);
	img->w = bmp.width;
	img->h = bmp.height;
	alloc_im(img);

	{
		/* scanlines are 32-bit aligned */
		int scanline_width = bmp.bmp_size / bmp.height / sizeof(u32);
		u32 *e = image + scanline_width * img->h;
		u32 *im = img->im;
		int d = (bmp.bpp == 24) ? 8 : bmp.bpp;
		u32 mask = (1<<d)-1;

		do {
			u32 *p = e - scanline_width;
			int w=0, bpp24 = 0;
			do {
				int s=32;
				u32 b;
				do {
					s -= d;
					b = be32(*p)>>s & mask;
					if (bmp.bpp == 24) {
						if (!bpp24) *im = 0;
						*im |= b << (bpp24 * d);
						if (++bpp24 == 3) {
							w++;
							if (*im == opt.alpha_key)
								*im = ALPHA;
							im++;
							bpp24 = 0;
						}
					}
					else {
						*im++ = img->pal[b];
						w++;
					}
				} while(s > 0 && w < img->w);
				p++;
			} while(p < e);
			e -= scanline_width;
		} while(e > image);
	}
	free(image);

	return;
}
