/*
 *	Copyright (C) 2001 Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include "toico.h"
#include <tiffio.h>

int istiff(u8 *p)
{
#ifdef TIFF_VERSION
	return *(u16*)p == TIFF_LITTLEENDIAN && *(u16*)(p+2) == le16(TIFF_VERSION)
		|| *(u16*)p == TIFF_BIGENDIAN && *(u16*)(p+2) == be16(TIFF_VERSION);
#else
	return *(u16*)p == TIFF_LITTLEENDIAN || *(u16*)p == TIFF_BIGENDIAN;
#endif
}

static tsize_t tread(thandle_t h, tdata_t d, tsize_t n)
{
	FILE *f = (FILE*)h;
	int v = fread(d, 1, n, f);
	return v;
}

static tsize_t twrite(thandle_t h, tdata_t d, tsize_t n)
{
	errx(1, "BUG: TIFF write called");
}

static int tclose(thandle_t h)
{
	return 0;
}

static toff_t tseek(thandle_t h, toff_t pos, int wh)
{
	FILE *f = (FILE*)h;
	fseek(f, pos, wh);
	return ftell(f);
}

static toff_t tsize(thandle_t h)
{
        FILE *f = (FILE*)h;
        toff_t size;
        fpos_t s;

        fgetpos(f, &s);
        fseek(f, 0, SEEK_END);
        size = ftell(f);
        fsetpos(f, &s);
        return size;
}

void readtiff(struct img *img, FILE *f, u8 sig[SIGN_SIZE])
{
	TIFF *tif;

	fseek(f, -SIGN_SIZE, SEEK_CUR);
	tif = TIFFClientOpen(img->name, "rm", (thandle_t)f,
		tread, twrite, tseek, tclose, tsize, 0, 0);
	if(!tif) goto badtiff;
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &img->w);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &img->h);
	alloc_im(img);
	TIFFReadRGBAImage(tif, img->w, img->h, (uint32*)img->im, 0);
	{
		u32 *p = img->im, *q = IMGEND(img);
		int w = img->w;
		do {
			u32 *e = p + w;
			q -= w;
			do {
				u32 v1 = *p, v2 = *q;
#define FIX(v) v&0xFF00FF00^0xFF000000 | v<<16 & 0xFF0000 | v>>16 & 0xFF
				*p++ = FIX(v2);
				*q++ = FIX(v1);
			} while(p < e);
			q -= w;
		} while(p < q);
	}
	TIFFClose(tif);
	return;
badtiff:
	errx(1, "%s: Can't open TIFF", img->name);
}
